#!/usr/bin/perl -w

## Copyright (c) 2004, Telecooperation Office (TecO),
## Universitaet Karlsruhe (TH), Germany.
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions
## are met:
##
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above
##     copyright notice, this list of conditions and the following
##     disclaimer in the documentation and/or other materials provided
##     with the distribution.
##   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
##     of its contributors may be used to endorse or promote products
##     derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


########################################
## lookup table mapping CCS codes to SDCC registers
## (this might be changed to use <register>,<bit>
## for improved performance...)

my %int_names = (	0xF2C0 => "GLOBAL",
			0xFFDD => "INT_LOWPRIO",  # this is invented by TecO convert tools
			0xF220 => "INT_RTCC",
			0x9D01 => "INT_TIMER1",
			0x9D02 => "INT_TIMER2",
			0xA002 => "INT_TIMER3",
			0xF210 => "INT_EXT",
			0xF008 => "INT_EXT1",
			0xF010 => "INT_EXT2",
			0xF020 => "INT_EXT3",
			0xF208 => "INT_RB",
			0x9D40 => "INT_AD",
			0x9D20 => "INT_RDA",
			0x9D10 => "INT_TBE",
			0x9D08 => "INT_SSP",
			0x9D04 => "INT_CCP1",
			0xA001 => "INT_CCP2",
			0xA008 => "INT_BUSCOL",
			0xA004 => "INT_LOWVOLT",
			0xA040 => "INT_COMP",
			0xA010 => "INT_EEPROM",
			0xA320 => "INT_RDA2",
			0xA310 => "INT_TBE2",
			0xA308 => "INT_TIMER4",
			0xA301 => "INT_CCP3",
			0xA302 => "INT_CCP4",
			0xA304 => "INT_CCP5"
		);

my %int_sources = (	0xF2C0 => "INTCONbits.GIE",
			0xFFDD => "INTCONbits.PEIE",
			0xF220 => "PIE1bits.TMR1IE",
			0x9D01 => "PIE1bits.TMR1IE",
			0x9D02 => "PIE1bits.TMR2IE",
			0xA002 => "PIE2bits.TMR3IE",
			0xF210 => "INTCONbits.INT0E",
			0xF008 => "INTCON3bits.INT1IE",
			0xF010 => "INTCON3bits.INT2IE",
			0xF020 => "INTCON3bits.INT3IE",
			0xF208 => "INTCONbits.RBIE",
			0x9D40 => "PIE1bits.ADIE",
			0x9D20 => "PIE1bits.RCIE",
			0x9D10 => "PIE1bits.TXIE",
			0x9D08 => "PIE1bits.SSPIE",
			0x9D04 => "PIE1bits.CCP1IE",
			0xA001 => "PIE2bits.CCP2IE",
			0xA008 => "PIE2bits.BCLIE",
			0xA004 => "PIE2bits.LVDIE",
			0xA040 => "PIE2bits.CMIE",
			0xA010 => "PIE2bits.EEIE",
			0xA320 => "PIE3bits.PC2IE",
			0xA310 => "PIE3bits.TX2IE",
			0xA308 => "PIE3bits.TMR4IE",
			0xA301 => "PIE3bits.CCP3IE",
			0xA302 => "PIE3bits.CCP4IE",
			0xA304 => "PIE3bits.CCP5IE"
		);

########################################
## emit file header

printf <<HEREDOC;
/* This file been automatically generated by $0.
 * You may modify this file if you wish, but any
 * modifications will be lost on subequent
 * recreation!
 */
#ifndef  CCS_COMPAT_H
#define  CCS_COMPAT_H

#ifdef SDCC_COMPAT_H
#error You included CCS_COMPAT_H as well as SDCC_COMPAT_H, which is invalid!
#else

#define  PIN2(pin)		CCS_PIN_##pin
#define  PIN(pin)		PIN2(pin)
#define  INT2(src)		CCS_INT_##src
#define  INT(src)		INT2(src)

HEREDOC

########################################
## provide macros to turn PORTx into TRISx or LATx

for my $i ("A","B","C","D","E","F","G","H","J") {
    printf "/* enable access to TRIS$i and LAT$i */\n";
    printf "#define TRISPORT$i TRIS$i\n";
    printf "#define LATPORT$i LAT$i\n";
    printf "#define PORTPORT$i PORT$i\n";
}
printf "\n";

########################################
# scan input for "sfr at <position> <name>"

%sfrs = ();

while (<>) {
	if (/sfr\s+at\s+((0x[0-9a-fA-F]+)|([0-9]+))\s+([a-zA-Z_][a-zA-Z_0-9]+)\b/i) {
		my $idx = $1;
		my $name = $4;
		$idx = oct($idx) if $idx =~ /^0/;

		$sfrs{$idx} = ${name};
#		$idx = $idx*8;
#		for ($i=0; $i<8; $i++) {
#			my $pin = $idx + $i;
#			printf "#define  CCS_PIN_%d\t${name}\n", ${pin};
#			printf "#define  CCS_PIN_0x%x\t${name}\n", ${pin};
#			printf "#define  CCS_PIN_0X%x\t${name}\n", ${pin};
#			printf "#define  CCS_PIN_0x%X\t${name}\n", ${pin};
#			printf "#define  CCS_PIN_0X%X\t${name}\n", ${pin};
#		}
	}
}

########################################
## emit macros to translate CCS pins into sfrs

for $i (sort keys %sfrs) {
    $name = $sfrs{$i};
    printf "/* sfr ${name}, register 0x%X */\n", $i;
    if ($name =~ /^PORT(.)$/) { $name = $1; }
    for ($j=0; $j < 8; $j++) {
	my $idx = 8 * $i + $j;
	printf "#define  CCS_PIN_%d\t${name},${j}\n", ${idx};
	#printf "#define  CCS_PIN_0x%x\t${name},${j}\n", ${idx};
	#printf "#define  CCS_PIN_0x%X\t${name},${j}\n", ${idx};
	#printf "#define  CCS_PIN_0X%x\t${name},${j}\n", ${idx};
	#printf "#define  CCS_PIN_0X%X\t${name},${j}\n", ${idx};
    }
}

########################################
## now prepare transcription of int sources

printf "\n";
printf "/* interrupt sources */\n";
for my $i (sort keys %int_sources) {
	printf ("#define  CCS_INT_" . $int_names{$i} . "\t" . $int_sources{$i} . "\n");
	printf ("#define  CCS_INT_0x%x\t" . $int_sources{$i} . "\n", $i);
	printf ("#define  CCS_INT_0X%x\t" . $int_sources{$i} . "\n", $i);
	printf ("#define  CCS_INT_0x%X\t" . $int_sources{$i} . "\n", $i);
	printf ("#define  CCS_INT_0X%X\t" . $int_sources{$i} . "\n", $i);
	printf ("#define  CCS_INT_%d\t" . $int_sources{$i} . "\n", $i);
}

########################################
## emit file footer

printf <<HEREDOC;

#endif // if only CCS_COMPAT_H
#endif // CCS_COMPAT_H
HEREDOC
