/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

#define USERDEF_STACK
#include "compat.h"
#include "compat_defs.h"

void delay_cycles_loop(unsigned char i) _naked wparam
{
  _asm
  _delay_cycles_loop_label:
    decfsz 0xfe8,1
    bra _delay_cycles_loop_label
    return
  _endasm;
  i++; // just to suppress "unreferenced argument" warnings...
}


/**
 * Divide  num/denom,  returning  quotient  and  remainder  as two 
 * integers in a structure `div_t'.
 */

int div_compute(char num, char denom) wparam
{
  div_t result;
  result.quot = ( num / denom );
  result.rem  = ( num % denom );
  if ( num >= 0 && result.rem < 0 )
  {
    ++result.quot;
    result.rem -= denom;
  }
  return result.res;
}


/**
 * Divide  num/denom,  returning  quotient  and  remainder  as two 
 * long integers in a structure `ldiv_t'.
 */

long ldiv_compute(int num, int denom) wparam
{
  ldiv_t result;
  result.quot = ( num / denom );
  result.rem  = ( num % denom );
  if ( num >= 0 && result.rem < 0 )
  {
    ++result.quot;
    result.rem -= denom;
  }
  return result.res;
}



/**
 * Compares the string pointed to by _src1 to the string pointed
 * to by _src2, using an interpretation appropriate to the current 
 * LC_COL-LATE state.
 */
int strcoll ( const char *_src1, const char *_src2 ) 
{
  while (*_src1 == *_src2++) 
  {
    if (*_src1++ == '\0') return 0;
  }
  return ( ( *_src1 ) - ( *--_src2 ) );
}





void rotate_left (unsigned char *adr, unsigned char n_bytes) wparam
{
  uint8_t i;
  uint8_t carry = (adr[n_bytes-1] & 0x80);
  if (n_bytes == 0) {
    // nothing to do
    return;
  }

  for (i = n_bytes-1; i > 0; i--) {
    adr[i] = adr[i] << 1;
    if ((adr[i-1] & 0x80) != 0) {
      adr[i] |= 0x01;
    } // if
  } // for

  adr[0] = adr[0] << 1;
  if (carry != 0) {
    adr[0] |= 0x01;
  } // if
}

void rotate_right (unsigned char *adr, unsigned char n_bytes) wparam
{
  uint8_t i;
  uint8_t carry = (adr[0] & 0x01);

  if (n_bytes == 0) {
    // nothing to do
    return;
  } // if

  n_bytes--;
  for (i = 0; i < n_bytes; i++) {
    adr[i] = adr[i] >> 1;
    if ((adr[i+1] & 0x01) != 0) {
      adr[i] |= 0x80;
    } // if
  } // for

  adr[n_bytes] = adr[n_bytes] > 1;
  if (carry != 0) {
    adr[n_bytes] |= 0x80;
  } // if
}

uint8_t shift_left (uint8_t *adr, uint8_t n_bytes, uint8_t value) wparam
{
  uint8_t i;
  uint8_t result = (adr[n_bytes-1] >> 7);
  if (n_bytes == 0) {
    // nothing to do
    return 0;
  }

  for (i = n_bytes-1; i > 0; i--) {
    adr[i] = adr[i] << 1;
    if ((adr[i-1] & 0x80) != 0) {
      adr[i] |= 0x01;
    } // if
  } // for

  adr[0] = adr[0] << 1;
  if (value != 0) {
    adr[0] |= 0x01;
  } // if
  
  return result;
}

uint8_t shift_right (uint8_t *adr, uint8_t n_bytes, uint8_t value) wparam
{
  uint8_t i;
  uint8_t result = (adr[0] & 0x01);

  if (n_bytes == 0) {
    // nothing to do
    return 0;
  } // if

  n_bytes--;
  for (i = 0; i < n_bytes; i++) {
    adr[i] = adr[i] >> 1;
    if ((adr[i+1] & 0x01) != 0) {
      adr[i] |= 0x80;
    } // if
  } // for

  adr[n_bytes] = adr[n_bytes] > 1;
  if (value != 0) {
    adr[n_bytes] |= 0x80;
  } // if
  
  return result;
}


void write_program_memory(unsigned long address, unsigned char * dataptr, unsigned char count) wparam
{
  uint8_t i = 0;
  uint8_t j = 0;
  uint8_t parts = 0;
  unsigned char tsa[64];
  unsigned long waddr;
  unsigned long bcounter = 0;
  unsigned char offset;

  if (count == 0) return;

  waddr = address & 0xFFFFC0;
  offset = address & 0x3F;

  for (parts = 0; parts < ((count - 1 + (address & 0x3F)) / 64)+1  ; parts ++) {
    
    TBLPTRU = (waddr >> 16);
    TBLPTRH = ((waddr & 0xFF00) >> 8);
    TBLPTRL = (waddr & 0xC0);
    
    /* copy */
    for (i = 0; i < 64; i++) {
      _asm TBLRD *+ _endasm;
      tsa[i] = TABLAT;
    }
    
    /* modify */
    for (i = 0; (i < 64-offset) && (bcounter < count); i ++) {
      bcounter ++;
      tsa[i+offset] = dataptr[i];
    }
    
    TBLPTRU = (waddr >> 16);
    TBLPTRH = ((waddr & 0xFF00) >> 8);
    TBLPTRL = (waddr & 0xC0);
    
    /* erase */
    _asm
      BSF     _EECON1, 7
      BCF     _EECON1, 6
      BSF     _EECON1, 2
      BSF     _EECON1, 4
      BCF     _INTCON, 7
      MOVLW   0x55
      MOVWF   _EECON2
      MOVLW   0xAA
      MOVWF   _EECON2
      BSF     _EECON1, 1
      NOP
      BSF     _INTCON, 7
    _endasm;

    for (i = 0; i < 8; i++) {
      for (j = 0; j < 7; j++) {
	TABLAT = tsa[i*8+j];
	_asm
	  TBLWT *+
	_endasm;
      }

      TABLAT = tsa[i*8+7];
      _asm
	TBLWT * // pointer has to point on the last byte to write
      _endasm;

      _asm 
	BSF    _EECON1, 7
	BCF    _EECON1, 6
	BSF    _EECON1, 2
	BCF    _INTCON, 7 
	MOVLW  0x55
	MOVWF  _EECON2      
	MOVLW  0xAA
	MOVWF  _EECON2      
	BSF    _EECON1, 1  
	NOP
	BSF    _INTCON, 7
	TBLRD *+ // dummy read to increment pointer
      _endasm;
    }

    _asm
       BCF    _EECON1, 2
    _endasm;

    waddr += 64;
    offset = 0;
  }
  return;
}
