/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

#ifndef COMPAT_DEFS_H
#define COMPAT_DEFS_H

/* output/input pin shortcut definitions to replace those of CCS */

#define PIN_A0  A, 0
#define PIN_A1  A, 1
#define PIN_A2  A, 2
#define PIN_A3  A, 3
#define PIN_A4  A, 4
#define PIN_A5  A, 5

#define PIN_B0  B, 0
#define PIN_B1  B, 1
#define PIN_B2  B, 2
#define PIN_B3  B, 3
#define PIN_B4  B, 4
#define PIN_B5  B, 5
#define PIN_B6  B, 6
#define PIN_B7  B, 7

#define PIN_C0  C, 0
#define PIN_C1  C, 1
#define PIN_C2  C, 2
#define PIN_C3  C, 3
#define PIN_C4  C, 4
#define PIN_C5  C, 5
#define PIN_C6  C, 6
#define PIN_C7  C, 7

#define PIN_D0  D, 0
#define PIN_D1  D, 1
#define PIN_D2  D, 2
#define PIN_D3  D, 3
#define PIN_D4  D, 4
#define PIN_D5  D, 5
#define PIN_D6  D, 6
#define PIN_D7  D, 7

#define PIN_E0  E, 0
#define PIN_E1  E, 1
#define PIN_E2  E, 2
#define PIN_E3  E, 3
#define PIN_E4  E, 4
#define PIN_E5  E, 5
#define PIN_E6  E, 6
#define PIN_E7  E, 7

#define PIN_F0  F, 0
#define PIN_F1  F, 1
#define PIN_F2  F, 2
#define PIN_F3  F, 3
#define PIN_F4  F, 4
#define PIN_F5  F, 5
#define PIN_F6  F, 6
#define PIN_F7  F, 7

#define PIN_G0  G, 0
#define PIN_G1  G, 1
#define PIN_G2  G, 2
#define PIN_G3  G, 3
#define PIN_G4  G, 4

#define ALL_PINS 0xFF


/* constants from CCS */
#define SPI_MASTER       0x20
#define SPI_SLAVE        0x24
#define SPI_L_TO_H       0
#define SPI_H_TO_L       0x10
#define SPI_CLK_DIV_4    0
#define SPI_CLK_DIV_16   1
#define SPI_CLK_DIV_64   2
#define SPI_CLK_T2       3
#define SPI_SS_DISABLED  1



/* CCS's interrupt levels (in SDCC fashion) */

#define GLOBAL					INTCONbits.GIE
#define INT_HIGHPRIO                            INTCONbits.GIE
#define INT_LOWPRIO                             INTCONbits.PEIE
#define INT_RTCC				INT_TIMER0
#define INT_TIMER0				INTCONbits.T0IE
#define INT_TIMER1				PIE1bits.TMR1IE
#define INT_TIMER2				PIE1bits.TMR2IE
#define INT_TIMER3				PIE2bits.TMR3IE
#define INT_EXT					INTCONbits.INT0E
#define INT_EXT1				INTCON3bits.INT1IE
#define INT_EXT2				INTCON3bits.INT2IE
#define INT_EXT3				INTCON3bits.INT3IE
#define INT_RB					INTCONbits.RBIE
#define INT_AD					PIE1bits.ADIE
#define INT_RDA					PIE1bits.RCIE
#define INT_TBE					PIE1bits.TXIE
#define INT_SSP					PIE1bits.SSPIE
#define INT_CCP1				PIE1bits.CCP1IE
#define INT_CCP2				PIE2bits.CCP2IE
#define INT_BUSCOL				PIE2bits.BCLIE
#define INT_LOWVOLT				PIE2bits.LVDIE
#define INT_COMP				PIE2bits.CMIE
#define INT_EEPROM				PIE2bits.EEIE
#define INT_RDA2				PIE3bits.PC2IE
#define INT_TBE2				PIE3bits.TX2IE
#define INT_TIMER4				PIE3bits.TMR4IE
#define INT_CCP3				PIE3bits.CCP3IE
#define INT_CCP4				PIE3bits.CCP4IE
#define INT_CCP5				PIE3bits.CCP5IE



/* Constants used for timer control */
/* - Timer 0 / RTCC */

#define RTCC_INTERNAL       0x00
#define RTCC_EXT_L_TO_H     0x20
#define RTCC_EXT_H_TO_L     0x30

#define RTCC_DIV_1      0x08
#define RTCC_DIV_2      0x00
#define RTCC_DIV_4      0x01
#define RTCC_DIV_8      0x02
#define RTCC_DIV_16     0x03
#define RTCC_DIV_32     0x04
#define RTCC_DIV_64     0x05
#define RTCC_DIV_128    0x06
#define RTCC_DIV_256    0x07

#define RTCC_ON         0x00            /* this is the default */
#define RTCC_OFF        0x80  

#define RTCC_8_BIT      0x40
#define RTCC_16_BIT     0x00            /* this is the default */


/* - Timer 1 */
#define T1_DISABLED         0x00
#define T1_INTERNAL         0x85        /* also enables 16 bit read/writes */
#define T1_EXTERNAL         0x87        /* also enables 16 bit read/writes */
#define T1_EXTERNAL_SYNC    0x83        /* also enables 16 bit read/writes */

#define T1_CLK_OUT          0x08

#define T1_DIV_BY_1         0x00
#define T1_DIV_BY_2         0x10
#define T1_DIV_BY_4         0x20
#define T1_DIV_BY_8         0x30

/* - Timer 2 */
#define T2_DISABLED         0x00
#define T2_DIV_BY_1         0x04
#define T2_DIV_BY_4         0x05
#define T2_DIV_BY_16        0x06

/* - Timer 3 */
#define T3_DISABLED         0x00
#define T3_INTERNAL         0x85        /* also enables 16 bit read/writes */
#define T3_EXTERNAL         0x87        /* also enables 16 bit read/writes */
#define T3_EXTERNAL_SYNC    0x83        /* also enables 16 bit read/writes */

#define T3_DIV_BY_1         0x00
#define T3_DIV_BY_2         0x10
#define T3_DIV_BY_4         0x20
#define T3_DIV_BY_8         0x30

/* - Timer 4 */
#define T4_DISABLED         0x00
#define T4_DIV_BY_1         0x04
#define T4_DIV_BY_4         0x05
#define T4_DIV_BY_16        0x06

/* - PSP - */
#define PSP_ENABLED      16
#define PSP_DISABLED      0

/* CCP */
#define CCP_OFF                         0
#define CCP_CAPTURE_FE                  4
#define CCP_CAPTURE_RE                  5
#define CCP_CAPTURE_DIV_4               6
#define CCP_CAPTURE_DIV_16              7
#define CCP_COMPARE_SET_ON_MATCH        8
#define CCP_COMPARE_CLR_ON_MATCH        9
#define CCP_COMPARE_INT                 0xA
#define CCP_COMPARE_INT_AND_TOGGLE      0x2       
#define CCP_COMPARE_RESET_TIMER         0xB
#define CCP_PWM                         0xC
#define CCP_PWM_PLUS_1                  0x1C
#define CCP_PWM_PLUS_2                  0x2C
#define CCP_PWM_PLUS_3                  0x3C
//#define CCP_USE_TIMER3                  0x100       ??



/* useful constants from CCS */

#define TRUE 1
#define FALSE 0

#define WDT_ON    1
#define WDT_OFF   0


/* additional types from CCS */

#define BYTE        unsigned char
#define byte        BYTE
#define BOOLEAN     unsigned char

#define int8        uint8_t
#define int16       uint16_t
#define int32       uint32_t

/* additional helpers */


/* div_t/ldiv_t structs for div/ldiv 
 * union workaround needed because SDCC can't use structs as fn result
 * REMEMBER: SDCC has 32bit longs, CCS has 16bit longs!
 */

typedef union {
  struct
    {
      char quot; /* Quotient.  */
      char rem;  /* Remainder.  */
    };
  int res;
  } div_t;

typedef union {
  struct
    {
      int quot; /* Quotient.  */
      int rem;  /* Remainder.  */
    };
  long res;
  } ldiv_t;

#endif
