/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

/*
 * Most code and ideas taken directy from MicroChip
 * http://ww1.microchip.com/downloads/en/DeviceDoc/i2c.pdf
 * and the corresponding processor data sheet
 */

#include <compat.h>

#undef I2C_SUFFIX
#define I2C_SUFFIX _hw

#undef i2c_wait
#undef i2c_poll
#undef i2c_start
#undef i2c_stop
#undef i2c_read
#undef i2c_write
#undef i2c_check

/*************************************************
 * Set up I2C-routines to use the hardware pins. *
 ************************************************/

//#define i2c_wait   /* defined below */
//#define i2c_poll   /* defined below */
#define i2c_start  _concat(i2c_start,I2C_SUFFIX)
//#define i2c_stop   /* defined below */
#define i2c_read   _concat(i2c_read ,I2C_SUFFIX)
#define i2c_write  _concat(i2c_write,I2C_SUFFIX)
#define i2c_check  _concat(i2c_check,I2C_SUFFIX)

/**************************************************
 * Implement the hardware supported I2C routines. *
 **************************************************/


/* 
 ***** currently not supported *******
#define i2cmode_speed_fast()            do {SSPSTATbits.SMP = 1; SSPADD = 11;} while(0)
#define i2cmode_speed_slow()            do {SSPSTATbits.SMP = 1; SSPADD = 11;} while(0)
#define i2cmode_speed()                 i2cmode_speed_slow()

#define i2cmode_master()		do {TRISDbits.TRISD4 = 1; TRISEbits.TRISE2 = 1; SSPCON1bits.SSPM0 = 0; SSPCON1bits.SSPM1 = 0; SSPCON1bits.SSPM2 = 0; SSPCON1bits.SSPM3 = 1; SSPCON1bits.SSPEN = 1;} while(0)
#define i2cmode_slave()			do {SSPCON1bits.SSPM2 = 1; SSPCON1bits.SSPM3 = 1; TRISDbits.TRISD4 = 1; TRISEbits.TRISE2 = 1;} while(0)
#define i2cmode()                       i2cmode_master()
*/

#define i2c_wait()                      do {while (!PIR1bits.SSPIF) {}; PIR1bits.SSPIF = 0;} while(0)
#define i2c_poll()                      SSPSTATbits.BF
#define i2c_stop()			do {PIR1bits.SSPIF = 0; SSPCON2bits.PEN = 1; i2c_wait(); SSPCON1bits.SSPEN = 0; PIR1bits.SSPIF = 0;} while(0)

#ifndef COMPAT_I2C_HW
#define COMPAT_I2C_HW

/* Prepare hardware I2C on pins C3 (SCL, clock) and C4 (SDA, data).
 * Make sure TRISC<4,3> is set up correctly to input or output! */
static void i2c_start() _naked {
  _asm
    bcf _PIR1, 3
    movlw 0x0B			// clock = F(osc) / (4 * WREG+1) (for master mode)
    movwf _SSPADD
    btfss _SSPCON1, 5		// check if I2C is already enabled
    bra _i2c_startcond
    bsf _SSPCON2, 1		// restart condition
    bra _rotate_i2c_start
_i2c_startcond:    
    movlw 0x28			// setup master mode I2C
    movwf _SSPCON1
    bsf _SSPCON2,0		// start condition
_rotate_i2c_start:
    btfss _PIR1, 3
    bra _rotate_i2c_start
    bcf _PIR1, 3		// reset interrupt flag
    return
  _endasm;
}

static uint8_t i2c_read(uint8_t ack) _naked wparam {
  _asm
    bsf _SSPCON2, 3		// enable I2C receive mode
    bsf _SSPCON2, 5		// default to NOT ACK
    tstfsz _WREG		// unless ack != 0...
    bcf _SSPCON2, 5		// ... in which case we shall send ACK
_rotate_i2c_read:		// wait for BUFFER FULL flag
    btfss _SSPSTAT, 0
    bra _rotate_i2c_read
    movf _SSPBUF, 0		// move the data into WREG as the return value
    btfss _SSPCON2, 5		// determine whether to send (N)ACK or not (maybe this line must be removed!)
    bsf _SSPCON2, 4		// initiate sending of ACK or NACK (always sends an ACK according to line above...)
    return
  _endasm;
  ack++; // just to suppress "unreferenced argument" warning
}

static uint8_t i2c_write(uint8_t content) _naked wparam {
  _asm
    bcf _PIR1, 3
    movwf _SSPBUF
_rotate_i2c_write:		// wait until data has been sent
    btfss _PIR1, 3
    bra _rotate_i2c_write
    bcf _PIR1, 3
    movlw 0x00			// return 0 on success
    btfsc _SSPCON2, 6		// check for ACK, skip if ACK was received
    movlw 0x01			// return 1 on error (no ACK from slave)
    return
  _endasm;
  content++; // just to suppress "unreferenced argument" warning
}

#endif // COMPAT_I2C_HW
