#!/usr/bin/perl -w

## Copyright (c) 2004, Telecooperation Office (TecO),
## Universitaet Karlsruhe (TH), Germany.
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions
## are met:
##
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above
##     copyright notice, this list of conditions and the following
##     disclaimer in the documentation and/or other materials provided
##     with the distribution.
##   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
##     of its contributors may be used to endorse or promote products
##     derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#########################################
###                                   ###
### G L O B A L E   V A R I A B L E N ###
###                                   ###
#########################################

my $found_fuses = 0;
my $int_flag = "";
my $last_line = "";
my $fop = 0;
my $uniquify = 1;

#######################################
###                                 ###
###   H I L F S - R O U T I N E N   ###
###                                 ###
#######################################

sub handle_fuses {
    # konvertiere #fuses
    if ($line =~ /^\s*\#(pragma\s+)?fuses/i) {
	if ($found_fuses == 0) {
	    $found_fuses = 1;
	    my %configs_default = ( "CONFIG1H" => 0x27,
				    "CONFIG2L" => 0x0F,
				    "CONFIG2H" => 0x0F,
				    "CONFIG3L" => 0x83,
				    "CONFIG3H" => 0x03,
				    "CONFIG4L" => 0x85,
				    "CONFIG5L" => 0xFF,
				    "CONFIG5H" => 0xC0,
				    "CONFIG6L" => 0xFF,
				    "CONFIG6H" => 0xE0,
				    "CONFIG7L" => 0xFF,
				    "CONFIG7H" => 0x40,
				    );
	    
	    my %configs = ();
	    
	    my %fuses_vars = (# brown out detection
			      "nobrownout" =>	"CONFIG2L",
			      "brownout" =>	"CONFIG2L",
			      # ???
			      "noprotect" =>	"CONFIG1H",
			      "protect" =>	"CONFIG1H",
			      # watchdog timer
			      "nowdt" =>	"CONFIG2H",
			      "wdt" =>		"CONFIG2H",
			      # watchdog timer postscaler
			      "wdt1" =>		"CONFIG2H",
			      "wdt2" =>		"CONFIG2H",
			      "wdt4" =>		"CONFIG2H",
			      "wdt8" =>		"CONFIG2H",
			      "wdt16" =>	"CONFIG2H",
			      "wdt32" =>	"CONFIG2H",
			      "wdt64" =>	"CONFIG2H",
			      "wdt128" =>	"CONFIG2H",
			      # power-up timer
			      "noput" =>	"CONFIG2L",
			      "put" =>		"CONFIG2L",
			      # low voltage program
			      "nolvp" =>	"CONFIG4L",
			      "lvp" =>		"CONFIG4L",
			      # oscillator to use
			      "lp" =>		"CONFIG1H",
			      "xt" =>		"CONFIG1H",
			      "hs" =>		"CONFIG1H",
			      "rc" =>		"CONFIG1H",
			      );
	    
	    my %fuses_flags = (# brown out detection
			       "nobrownout" =>  "_BODEN_OFF_2L",
			       "brownout" =>    "_BODEN_ON_2L",
			       # ???
			       "noprotect" =>   "",
			       "protect" =>     "",
			       # watchdog timer
			       "nowdt" =>       "_WDT_OFF_2H",
			       "wdt" =>         "_WDT_ON_2H",
			       # watchdog timer postscaler
			       "wdt1" =>        "_WDTPS_1_1_2H",
			       "wdt2" =>        "_WDTPS_1_2_2H",
			       "wdt4" =>        "_WDTPS_1_4_2H",
			       "wdt8" =>        "_WDTPS_1_8_2H",
			       "wdt16" =>       "_WDTPS_1_16_2H",
			       "wdt32" =>       "_WDTPS_1_32_2H",
			       "wdt64" =>       "_WDTPS_1_64_2H",
			       "wdt128" =>      "_WDTPS_1_128_2H",
			       # power-up timer
			       "noput" =>       "_PUT_OFF_2L",
			       "put" =>         "_PUT_ON_2L",
			       # low voltage program
			       "nolvp" =>       "_LVP_OFF_4L",
			       "lvp" =>         "_LVP_ON_4L",
			       # oscillator to use
			       "lp" =>          "_OSC_LP_1H",
			       "xt" =>          "_OSC_XT_1H",
			       "hs" =>          "_OSC_HS_1H",
			       "rc" =>          "_OSC_RC_1H",
			       );
	    my $warnings = "";
	    foreach $fuse (keys %fuses_vars) {
		if ($line =~ /\b$fuse\b/i) {
		    $line =~ s/\b$fuse\b//i;
		    if ($fuses_flags{$fuse} ne "") {
			$configs{$fuses_vars{$fuse}} .= " & " . $fuses_flags{$fuse};
		    } else {
			$warnings .= "\"$fuse\" ignored! ";
		    }
		} else {
		    #printf ("not found: " . $fuse . "\n");
		}
	    }
	    $line =~ s/(\#(pragma\s+)?fuses\s+)//i;
	    $line =~ s/[\r\n\t, ]+//ig;
	    if ($line ne "") {
		$warnings .= "\"$line\" ignored! ";
		$line = "";
	    }
	    # emit config words in SDCC syntax
	    foreach $conf (sort keys %configs) {
		$last_line .= sprintf ("code char at __$conf " . lc ($conf) . " = 0xFF $configs{$conf};\n");
	    }
	    if ($warnings ne "") {
		$last_line .= sprintf ("// WARNING: $warnings");
	    }
	} else {
	    $last_line .= sprintf ("#warning More than one #fuses directive found --- ignored!\n");
	    $line = "//" . $line;
	} # if
    } # if ($line =~ /\bfuses\b/)
} # sub handle_fuses



sub handle_iomodes {
    # behandle IO-Modi
    $line =~ s/\#(pragma\s+)?use\s+fast_io\((.)\).*$/\#undef set_mode_$2\n\#define set_mode_$2(input, pin) portmode_fast($2, input, pin) \n/ig;
    if ($line =~ /\#(pragma\s+)?use\s+fixed_io\((.)_outputs=(.*)\).*$/) {
	$line =~ s/\#(pragma\s+)?use\s+fixed_io\((.)_outputs=(.*)\).*$/\#undef set_mode_$2\n\#define set_mode_$2(input, pin) portmode_fixed($2, input, pin) \n/ig; 
	$pins = $3;
	$port = uc($2);
	$mask = 0xFF;
	while ($pins =~ /PIN_$port(.)/i) {
	    $mask -= (1 << $1);
	    $pins =~ s/PIN_$port(.)//i;
	}
	$last_line .= sprintf("#undef FIXEDMASK_TRIS".$port."\n");
	$last_line .= sprintf("#define FIXEDMASK_TRIS".$port." 0x%02X\n", $mask);
    }
    $line =~ s/\#(pragma\s+)?use\s+standard_io\((.)\).*$/\#undef set_mode_$2\n\#define set_mode_$2(input, pin) portmode_standard($2, input, pin)\n/ig;
    if ($line =~ /(\#define |\#undef )set_mode_(.)/i) {
	$lc = lc($2);
	$uc = uc($2);
	$line =~ s/mode_$lc/mode_$uc/ig;
	$line =~ s/$lc, input/$uc, input/ig;
    }
} # sub handle_iomodes

sub handle_i2cmodes {
    # behandle I2C-Modi
    if ($line =~ /\#(pragma\s+)?use\s+i2c\s*\((.+)\).*$/i) {
	$params = $2;
	if ($params =~ /\bmaster\b/i) {
	    # that's just fine...
	} else {
	    # I2C slave mode not implemented
	    $last_line .= "#warning I2C slave mode not supported!\n";
	}
	
	# extrahiere die CLOCK und DATA PINs
	$last_line .= "#undef I2C_SCL\n";
	$last_line .= "#undef I2C_SDA\n";
	$last_line .= "#undef I2C_SUFFIX\n";
	my $i2c_scl="";
	my $i2c_sda="";
	
	if ($params =~ /\bsda\s*=\s*(\w+)\b/i) {
	    $last_line.= "#define I2C_SDA   \t$1\n";
	    $i2c_sda = $1;
	}

	if ($params =~ /\bscl\s*=\s*(\w+)\b/i) {
	    $last_line .= "#define I2C_SCL   \t$1\n";
	    $i2c_scl = $1;
	}
	
	# definiere eindeutiges Suffix f&uuml;r die neuen Software I2C-Routinen
	if ($i2c_scl ne "" && $i2c_sda ne "") {
		$last_line .= sprintf ("#define I2C_SUFFIX\t_%s_%s\n", $i2c_scl, $i2c_sda);
	} else {
		$last_line .= sprintf ("#define I2C_SUFFIX\t_%d_%d_%d\n", getppid(), time(), $uniquify++);
	}

	# erzeuge die Methoden
	if ($params =~ /\bFORCE_HW\b/i) {
	    # Hardware-I2C-Routinen verwenden (geht nur mit den Hardware PINs C3/C4)!
	    $line = "#include <compat_i2c.h>\n";
	} else {
	    # Software-I2C-Routinen generieren
	    $line = "#include <software_i2c.c>\n";
	}
    }
} # sub handle_i2cmodes



sub handle_binary_literals {
    # ersetze 0b???????? durch 0x!!
    while ($line =~ /\b(0b[01]{8})\b/i) {
	$val = $1;
	$dec = $val;
	$dec = oct($dec) if $dec =~ /^0/;
	$hex = sprintf("0x%02X", $dec);
	$line =~ s/$val/$hex/ig;
    } # while
} # sub handle_binary_literals



#######################################
###                                 ###
###   H A U P T - P R O G R A M M   ###
###                                 ###
#######################################

# TODO: add all PICC-headers here...
# These headers are TecO-headers!
#my @PICC_HEADERS = ();
#my @PICs = ("18f452", "18f6620", "18f6720");
#my @suffices = ("");
#foreach $pic (@PICs) {
#  foreach $suffix (@suffices) {
#    @PICC_HEADERS = (@PICC_HEADERS, "${pic}${suffix}.h");
#  }
#  #$pic =~ s/f/F/;
#  #@PICC_HEADERS = (@PICC_HEADERS, "${pic}_R.h");
#}
#printf "// CCS headers to be removed:\n";
#foreach $header (@PICC_HEADERS) {
#  print "// $header\n";
#}
#printf "// END of CCS headers\n";

# binde die Kompatibilit&auml;ts-Schicht automatisch ein
$last_line .= "#include <compat.h>\n";

while ($line = <>) {
	# bilde SFR-Definitionen auf SDCCs SFRs ab
	$line =~ s/\#(pragma\s+)?byte\s+PIC_(\w+)\s*=\s*\w+/\#define PIC_$2\t$2/ig;

#	$line =~ s/\#define\s+PIC_(\w+)\s*\w+$/\#define PIC_$1\t$1/ig;

	# ersetze #byte
	## pin variable to address
	$line =~ s/\#(pragma\s+)?byte\s+(\w+)\s*=\s*([1-9]\d*|0([xX][0-9a-fA-F]+|[0-7]*))/uint8_t at $3 $2;/ig;

	## pin variable to overlay with another variable
	$line =~ s/\#(pragma\s+)?byte\s+(\w+)\s*=\s*(\w+)/uint8_t at $3 $2;/ig;

	# ersetze #bit
	$line =~ s/\#(pragma\s+)?bit\b/\/\/#pragma bit/ig;
	

	# wandle Pfadnamen mit Backslash zu Slashes um
	if ($line =~ /\#include(\s+)(["<])([^">]+)([">])(.*)/i) { # closing "])
	    $line =~ s#\\#/#g;
	}

	# behandle #fuses
	handle_fuses;

	# behandle #rom
	if ($line =~ /\#(pragma\s+)?rom\s+/i) {
	    $line =~ s/\#(pragma\s+)?rom\s+(\w+)\s*=/code uint16_t at $2 __pos$2\[\] =/i;
	    $fop = 1;
	}
	if (($fop == 1) && ($line =~ /\}/)) {
	    $line =~ s/\}/\}\;/;
	    $fop = 0;
	}


	# Gro&szlig;-Kleinschreibung anpassen
	$line =~ s/\#if\b/\#if/i;
	$line =~ s/\#else\b/\#else/i;
	$line =~ s/\#endif/\#endif/i;
	$line =~ s/\#define\b/\#define/i;
	$line =~ s/\#undef\b/\#undef/i;
	$line =~ s/\#pragma\b/\#pragma/i;
	
	# behandle inline assembler
	$line =~ s/\#(pragma\s+)?asm/_asm/ig;
	$line =~ s/\#(pragma\s+)?endasm/_endasm\;/ig;

	# Funktionen als inline oder (explizit) nicht inline markieren
	## either turn directive into pragma...
	$line =~ s/\#(pragma\s+)?inline\b/\/\/\#pragma inline/ig;
	#$line =~ s/\#(pragma\s+)?separate\b/\/\/\#pragma separate/ig;

	## ... or ignore it completely (as long as SDCC cannot do inlining)
	#$line =~ s/\#(pragma\s+)?inline\b//ig;
	$line =~ s/\#(pragma\s+)?separate\b//ig;

	# suche Interrupthandler-Routinen
	if ($line =~ /\#(pragma\s+)?int_(\w+)/i) {
	    $int_flag = lc($2);
	    if ($line =~ /\bfast\b/i) {
		$last_line .= sprintf("#define ". $int_flag . "_fast\n");
	    }
	    if ($line =~ /\bnoclear\b/i) {
		$last_line .= sprintf("#define " . $int_flag . "_noclear\n");
	    }
	    $line = "";
	}
	
	# ermittle auf Interrupthandler-Direktive folgende Funktionsdefinition
	if (($line =~ /(\w+)\s*\((.*)$/) && ($int_flag ne "")) {
	    $last_line .= sprintf ("#define " . $int_flag . "_handler " . $1 . "\n");
	    $line = "void $1\($2\n";
	    $int_flag = "";
	}
	
	# behandle IO-Modi
	handle_iomodes;

	# behandle I2C-Modi
	handle_i2cmodes;

	# USE DELAY
        if ($line =~ /\#(pragma\s+)?use\s+delay\s*\(\w+=(\d+)/i) {
	    $last_line .= sprintf("#undef RESTART_WDT\n");
            if ($line =~ /RESTART_WDT/i) {
                $last_line .= sprintf("#define RESTART_WDT 1\n");
            }
            $speed = $2;

            $line =~ s/\#(pragma\s+)?use\s+delay.*/\#define CLOCK_SPEED $speed\n\#include \<delay.c\>/i;
        }
     	$line =~ s/\bdelay_us\((1?\d)\)/delay_us_short\($1\)/ig;

	# nicht zu &uuml;bersetzende Direktiven verwerfen
	$line =~ s/\#(pragma\s+)?(use|case|nolist|list|device|org|type|id|ignore_warnings|priority|zero_ram|opt|int_default|int_global).*$//ig;

	# reserve-Pragma behandeln
	$line =~ s/\#(pragma\s+)?reserve\s+(.*)/\#pragma reserve $2/i;
	
	while ($line =~ /\#pragma\s+reserve\s+(\w+)\s+,(.*)/i) {
	    $last_line .= sprintf("#pragma reserve ". $1 . "\n");
	    $line =~ s/\#pragma\s+reserve\s+(\w+),(.*)$/\#pragma reserve $2/i;
	}

	# locate-Pragma behandeln
	if ($line =~ /\#(pragma\s+)?locate\s+(\w+)\s*=\s*(\w+)/i) {
	    $addr = $3;
	    if ($last_line =~ /^[^;]+=/) {
		$last_line .= sprintf ("#warning locate-pragma may not fully supported\n");
		$last_line .= sprintf ("#error Unable to use initialized data on fixed positions!\n");
	    } else {
		$last_line =~ s/\b(\w+)\s*;/at $addr $1;/;
		if ($last_line =~ /\b(u?int8_t|u?int16_t|u?int32_t|char|byte)\b/) {
		    $type = $1;
		    $line = "#pragma reserve ". $addr;
		    $addr = oct($addr) if $addr =~ /^0/;
		    if ($type =~ /u?int16_t/) {
			$line .= "\n#pragma reserve " . sprintf ("0x%02X", 1+$addr);
		    }
		    if ($type =~ /u?int32_t/) {
			$line .= "\n#pragma reserve " . sprintf ("0x%02X", 1+$addr);
			$line .= "\n#pragma reserve " . sprintf ("0x%02X", 2+$addr);
			$line .= "\n#pragma reserve " . sprintf ("0x%02X", 3+$addr);
		    }
		    $last_line .= sprintf ("#warning locate-pragma may not fully supported --- check manually\n");
		} else {
		    $last_line .= sprintf ("#warning locate-pragma may not fully supported\n");
		    $last_line .= sprintf ("#error Unable to determine the amount of space to reserve (unknown type)!\n");
		}
	    }
	}

	# Polymorphie-Aufl&ouml;sung
	# Probleme: Funktionsaufrufe als Argument
	#$line =~ s/\bbit_(set|clear|test)(\s*)\(((\w|\s)+)\)/bit_${1}1$2($3)/g;
	$line =~ s/\bi2c_read(\s*)\((\s*)\)/i2c_read$1\(1\)/g;
	$line =~ s/\bread_adc\s*\((\s*)\)/read_adc_mode\($1\)/g;
	$line =~ s/\bspi_read\s*\((\s*)\)/spi_read_data\($1\)/g;
	
	# Passe Datentypen CCS --> SDCC an
	# CCS: (char|short|int|long) being unsigned with (8|1|8|16) bit
	# SDCC: (char|short|int|long) being signed with (8|16|16|32) bit
        # ACHTUNG: short (aka bit) werden nicht behandelt!
	$line =~ s/\blong\s+int\b/long/g;
	$line =~ s/\bshort\s+int\b/short/g;

	# replace signed types
	$line =~ s/\bsigned\s+(char)\b/char/g;
	$line =~ s/\bsigned\s+(short|int)\b/int8_t/g;
	$line =~ s/\bsigned\s+int(8|16|32)\b/int$1_t/g;
	$line =~ s/\bsigned\s+(long)\b/int16_t/g;

	# replace unsigned types
	$line =~ s/\b(unsigned\s+)?(char)\b/unsigned char/g;
	$line =~ s/\b(unsigned\s+)?(short|int)\b/uint8_t/g;
	$line =~ s/\b(unsigned\s+)?int(8|16|32)\b/uint$2_t/g;
	$line =~ s/\b(unsigned\s+)?(long)\b/uint16_t/g;


	# Ersetzt Bin&auml;r-Literale durch Hex-Literale
	handle_binary_literals;

	# ersetze/streiche einige INCLUDEs (alle PICC-Header muessen raus!)
#	foreach $header (@PICC_HEADERS) {
#	    printf ("Checking for header $header\n");
#	    if ($line =~ /\#include\s*["<](.*[\/\\])*$header[">]/) {
#		$line = "// old CCS header: $line";
#	    }
#	}

    print $last_line;
    $last_line = $line;
}
print $last_line;
print "\n";
