#!/usr/bin/perl -w

## Copyright (c) 2004, Telecooperation Office (TecO),
## Universitaet Karlsruhe (TH), Germany.
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions
## are met:
##
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above
##     copyright notice, this list of conditions and the following
##     disclaimer in the documentation and/or other materials provided
##     with the distribution.
##   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
##     of its contributors may be used to endorse or promote products
##     derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO: implement #priority pragma and sort @int_sources accordingly


my @int_sources = ( "ad",
#		    "adof",
		    "buscol",
#		    "button",
		    "ccp1", "ccp2", "comp", "eeprom", "ext", "ext1", "ext2", "ext3",
#		    "i2c", "lcd",
		    "lowvolt", "psp", "rb", "rc", "rda", "rtcc", "ssp", "tbe", "timer0",
		    "timer1", "timer2", "timer3", "timer4"
		    );

my %int_bits = ( "ad" => "PIR1bits.ADIF",
#		 "adof" => "",
		 "buscol" => "PIR2bits.BCLIF",
#		 "button" => "",
		 "ccp1" => "PIR1bits.CCP1IF",
		 "ccp2" => "PIR2bits.CCP2IF",
		 "comp" => "PIR2bits.CMIF",
		 "eeprom" => "PIR2bits.EEIF",
		 "ext" => "INTCONbits.INT0F",
		 "ext1" => "INTCON3bits.INT1F",
		 "ext2" => "INTCON3bits.INT2F",
		 "ext3" => "INTCON3bits.INT3F",
#		 "i2c" => "",
#		 "lcd" => "",
		 "lowvolt" => "PIR2bits.LVDIF",
		 "psp" => "PIR1bits.PSPIF",
		 "rb" => "INTCONbits.RBIF",
		 "rc" => "PIR1bits.RCIF",
		 "rda" => "PIR1bits.RCIF",
		 "rtcc" => "INTCONbits.T0IF",
		 "ssp" => "PIR1bits.SSPIF",
		 "tbe" => "PIR1bits.TXIF",
		 "timer0" => "INTCONbits.T0IF",
		 "timer1" => "PIR1bits.TMR1IF",
		 "timer2" => "PIR1bits.TMR2IF",
		 "timer3" => "PIR2bits.TMR3IF",
		 "timer4" => "PIR3bits.TMR4IF",
);

# some additional registers (WREG, STATUS and BSR are saved automatically) have to be saved
my @saved_regs = ( );

########################################
##
## subroutines to save and restore
## registers on interrupt entry
##
########################################

sub save_regs () {
  if (@saved_regs > 0) {
    printf "  _asm\n";
    for ($i=0; $i < @saved_regs; $i++) {
      printf "    MOVFF\t" . $saved_regs[$i] . ", POSTDEC1\n";
    }
    printf "  _endasm;\n";
    printf "\n";
  } # if
}

sub restore_regs () {
  if (@saved_regs > 0) {
    printf "  _asm\n";
    for ($i=@saved_regs-1; $i >= 0; $i--) {
      printf ("    MOVFF\tPREINC1, " . $saved_regs[$i] . "\n");
    }
    printf ("  _endasm;\n");
  } # if
}

########################################
##
## create file header
##
########################################

printf <<HEREDOC;
/* This file has automatically been generated by $0.
 * You are free to modify this file, but any changes
 * made will be lost upon recreation via $0.
 */
#ifndef __DISPTAB_C__
#define __DISPTAB_C__

#define NO_SDCC_STACK 1
#include <tools.h>
#include <initprio.c>

HEREDOC

########################################
##
## create dispatch table (low priority)
##
########################################

printf <<HEREDOC;
/***********************************************
 * Dispatch table for low priority interrupts. *
 ***********************************************/
void __irqhandler_low () interrupt 2 {
  disable_interrupts(INT_LOWPRIO);

HEREDOC
save_regs();
for $irq (@int_sources) {
    printf "#if defined (${irq}_handler) && !defined (${irq}_fast)\n";
    printf "  if (" . $int_bits{$irq} . ") {\n";
    printf "    ${irq}_handler();\n";
    printf "#ifndef ${irq}_noclear\n";
    printf "    " . $int_bits{$irq} . " = 0;\n";
    printf "#endif\n";
    printf "  }\n";
    printf "#endif\n";
    printf "\n";
}
restore_regs();
printf <<HEREDOC;
  enable_interrupts(INT_LOWPRIO);

}

HEREDOC

########################################
##
## create dispatch table (high priority)
##
########################################

printf <<HEREDOC;
/************************************************
 * Dispatch table for high priority interrupts. *
 ************************************************/
void __irqhandler_high () interrupt 1 {
  disable_interrupts(GLOBAL);

HEREDOC
save_regs();
for $irq (@int_sources) {
    printf "#if defined (${irq}_handler) && defined (${irq}_fast)\n";
    printf "  if (" . $int_bits{$irq} . ") {\n";
    printf "    ${irq}_handler();\n";
    printf "#ifndef ${irq}_noclear\n";
    printf "    " . $int_bits{$irq} . " = 0;\n";
    printf "#endif\n";
    printf "  }\n";
    printf "#endif\n";
    printf "\n";
}
restore_regs();
printf <<HEREDOC;
  enable_interrupts (GLOBAL);
}

HEREDOC

########################################
##
## create file footer
##
########################################

printf <<HEREDOC;
#else
  #warning Interrupt dispatch table has been included more than once!
#endif // __DISPTAB_C__
HEREDOC
