/* -*- text-mode -*- */

%{
/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

  #include <stdio.h>
  #include <string.h>

  static int inInlineAsm = 0;

  static int isTargetKeyword(char *s);
  static int isAssemblerDirective(char *s);
  static isOpcode (const char *id);
%}

%%
"\#asm"       { isAssemblerDirective (yytext); printf ("%s", yytext); }
"\#endasm"    { isAssemblerDirective (yytext); printf ("%s", yytext); }
  /* prevent #bit from being transformed into #__PIC_bit below */
"#bit"        { printf ("%s", yytext); }

  /* Kommentare */
\/\/.*$                     { printf ("%s", yytext); }
\/\*([^\*]|\n|\*[^/])*\*+\/ { printf ("%s", yytext); }

  /* Strings und Zeichen */
\"([^"]|\\\")*\"            { printf ("%s", yytext); /* " */ }
\'([^']|\\\')*\'            { printf ("%s", yytext); /* ' */ }

  /* integer Literale */
(([1-9][0-9]+)|(0[xX][0-9a-fA-F]+)|(0[1-7]+)) { printf ("%s", yytext); }

  /* Bezeichner */
[a-zA-Z_][a-zA-Z0-9_]*  {
	 if (isTargetKeyword (yytext)) {
	   printf ("__PIC_%s", yytext);
	 } else {
	   if (inInlineAsm && !isOpcode(yytext)) {
		 // ersetze registername durch "_[registername]" OHNE f&uuml;hrendes PIC
		 printf ("ASM(%s)", yytext);
	   } else {
		 printf ("%s", yytext);
	   }
	 }
   }

%%

int main (int args, char ** argv) {
//  printf ("Lexer gestartet!\n");
  yylex();
//  printf ("Lexer fertig?!?\n");
}

static char *_pic16_keywords[] =
{
  "at",
  "bit",
  "code",
  "critical",
//  "register",  /* keyword for CCS as well */
  "data",
  "far",
  "idata",
  "interrupt",
  "near",
  "pdata",
  "reentrant",
  "sfr",
  "sbit",
  "using",
  "xdata",
  "_data",
  "_code",
  "_generic",
  "_near",
  "_xdata",
  "_pdata",
  "_idata",
  "_naked",
  NULL
};

static int isTargetKeyword(char *s)
{
  int i;

  for (i = 0 ; _pic16_keywords[i] ; i++ ) {
    if (strcmp(_pic16_keywords[i],s) == 0)
      return 1;
  }

    return 0;
}

static int isAssemblerDirective (char *text) {
  if (strcmp (text, "#asm") == 0) {
    // Anfang eines inline assembler-Blocks
    inInlineAsm = 1;
    return 1;
  } else if (strcmp (text, "#endasm") == 0) {
    // Ende des inline assembler-Blocks
    inInlineAsm = 0;
    return 1;
  }
  // keine inline assembler Anweisung
  return 0;
}

static const char * const mnemonics[] = {
  "addwf", "addwfc", "andwf",
  "clrf", "comf", "cpfseq", "cpfsgt", "cpfslt",
  "decf", "decfsz", "dcfsnz",
  "incf", "incfsz", "infsnz", "iorwf",
  "movf", "movff", "movwf", "mulwf",
  "negf",
  "rlcf", "rlncf", "rrcf", "rrncf",
  "setf",
  "subfwb", "subwf", "subwfb", "swapf",
  "tstfsz",
  "xorwf",

  "bcf", "bsf", "btfsc", "btfss", "btg",
  "bc", "bn", "bov", "bz", 
  "bnc", "bnn", "bnov", "bnz", 
  "bra", "call", "clrwdt",
  "daw",
  "goto",
  "nop",
  "pop", "push",
  "rcall", "reset", "retfie" "retlw", "return",
  "sleep"

  "addlw", "andlw", "iorlw",
  "lfsr",
  "movlb", "movlw", "mullw",
  "retlw",
  "sublw",
  "xorlw",

  "tblrd", "tblwt",

  "w", "f"   // destination identifiers
};

static int isOpcode (const char *id) {
  int i=0;
  while (mnemonics[i]) {
    if (strcasecmp (id, mnemonics[i]) == 0)
      // mnemonic found
      return 1;
    i++;
  } // while
  return 0;
}
