/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/


/*
 * Code parts and ideas taken from
 * - I2C low level routines for Intel 8xC51Fx chips from Thorsten Godau
 * - MicroChip Processors Data Sheet
 * - code found on www.sprut.de (Joerg Bredendiek, we guess)
 * Thanks!
 */

#include <software_i2c.h>

#ifndef I2C_SCL
#error I2C_SCL undefined while including software_i2c.c
#define I2C_SCL PIN_C3
#endif

#ifndef I2C_SDA
#error I2C_SDA undefined while including software_i2c.c
#define I2C_SDA PIN_C4
#endif

#ifndef I2C_SUFFIX
#error I2C_SUFFIX undefined while including software_i2c.c
#define I2C_SUFFIX _concat(I2C_SCL,I2C_SDA)
#endif

/************************************************
 * Set up I2C-routines to use the desired pins. *
 ************************************************/

#undef i2c_wait
#undef i2c_poll
#undef i2c_start
#undef i2c_stop
#undef i2c_read
#undef i2c_write
#undef i2c_check

#define i2c_start  _concat(i2c_start,I2C_SUFFIX)
#define i2c_stop   _concat(i2c_stop ,I2C_SUFFIX)
#define i2c_read   _concat(i2c_read ,I2C_SUFFIX)
#define i2c_write  _concat(i2c_write,I2C_SUFFIX)
#define i2c_check  _concat(i2c_check,I2C_SUFFIX)

/************************************************************
 * Implement the software I2C routines on the desired pins. *
 ************************************************************/

static void i2c_start ()
{
  set_pin_to_output(I2C_SCL);
  set_pin_to_output(I2C_SDA);
  output_high(I2C_SDA);
  output_high(I2C_SCL);
  delay_cycles(25);
  output_low(I2C_SDA);
  delay_cycles(25);
  output_low(I2C_SCL);
  return;
}

static void i2c_stop ()
{
  output_low(I2C_SDA);
  delay_cycles(25);
  output_high(I2C_SCL);
  delay_cycles(25);
  output_high(I2C_SDA);
  return;
}

static uint8_t i2c_check ()
{
  uint8_t timeout = 50;

  output_high(I2C_SDA);
  output_high(I2C_SCL);
  output_float(I2C_SDA); // new
  while(!input(I2C_SCL));
  delay_cycles(25);

  while (timeout-- && input(I2C_SDA)) {
    // wait for ack...
  }

  if (input(I2C_SDA))
    {
      output_low(I2C_SCL);
      return 1;
    }

  output_low(I2C_SCL);
  output_high(I2C_SDA);
  set_pin_to_output(I2C_SDA);
  return 0;
}

static uint8_t i2c_read (uint8_t ack) wparam
{
  uint8_t counter = 8;
  uint8_t result = 0;
  
  for (; counter > 0; counter--) 
  {
    output_float(I2C_SDA);
    output_high(I2C_SCL);
    while(!input(I2C_SCL));
    delay_cycles(25);
    result <<= 1;
    if (input(I2C_SDA))
      result |= 1;
    output_low(I2C_SCL);
    delay_cycles(25);
  }

  if (ack) // may work ... 
  {
    set_pin_to_output(I2C_SDA);
    output_low(I2C_SDA);
    output_high(I2C_SCL);
    delay_cycles(25);
    output_low(I2C_SCL);
  }
  return result;  
}

static uint8_t i2c_write (uint8_t content) wparam
{
  uint8_t counter = 8;
  for (; counter > 0; counter--)
    {
      if (content & 0x080)
	{
	  output_high(I2C_SDA);
	}
      else
	{
	  output_low(I2C_SDA);
	}
      output_high(I2C_SCL);
      while (!input(I2C_SCL));
      delay_cycles(25);
      output_low(I2C_SCL);
      delay_cycles(25);
      content <<=1;
    }
  return(i2c_check());
}

