/**
  * Copyright (c) 2004, Telecooperation Office (TecO),
  * Universitaet Karlsruhe (TH), Germany.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
  *   * Neither the name of the Universitaet Karlsruhe (TH) nor the names
  *     of its contributors may be used to endorse or promote products
  *     derived from this software without specific prior written
  *     permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

#include <compat.h>

#define set_pin_to_output(port,pin)	TRIS##port &= 0xFF ^ (1U << pin)
#define set_pin_to_output1(pin)		set_pin_to_output(pin)

#define UART_CLK		A,1
#define UART_DATA		A,2

#define PORT2(port,pin)		PORT##port,pin
#define PORT(pin)		PORT2(pin)
#define UART_CLK_PORT		PORT(UART_CLK)
#define UART_DATA_PORT		PORT(UART_DATA)

#define UARTWaitForRisingClk	{ while(bit_test1(UART_CLK_PORT)); while(!bit_test1(UART_CLK_PORT)); }
#define UARTWaitForFallingClk	{ while(!bit_test1(UART_CLK_PORT)); while(bit_test1(UART_CLK_PORT)); }

void UARTInit (uint8_t isSender)
{
	UART_CLK_PORT;
	// CLK is input (start external clock generation after this
	// as it might set this pin to out again...)
	output_float(UART_CLK);

	if (isSender) {
		// DATA is output
		set_pin_to_output1(UART_DATA);
	} else {
		// DATA is input
		output_float(UART_DATA);
	}
}

/* Emit count bytes from val on the serial IO pins. */
void UARTWrite (uint8_t *val, uint8_t count)
{
	uint8_t bitcount = 8;
	uint8_t buf;

	while (count--) {
	  // send start bit
	  UARTWaitForRisingClk;
	  output_low(UART_DATA);

	  buf = *val;
	  val++;

	  for (bitcount=8; bitcount > 0; bitcount--) {
	    // emit next bit after rising clock edge (LSB first)
	    if (bit_test(buf,0)) {
	      UARTWaitForRisingClk;
	      output_high(UART_DATA);
	    } else {
	      UARTWaitForRisingClk;
	      output_low(UART_DATA);
	    }
	    buf >>= 1;
	  } // for i

	  // send stop bit
	  UARTWaitForRisingClk;
	  output_high(UART_DATA);
	} // while
}

/* Receive count bytes into val from the serial IO pins.
 * Returns the number of bytes successfully read. */
uint8_t UARTRead (uint8_t *val, uint8_t count)
{
	uint8_t bitcount;
	uint8_t readbytes=0;
	uint8_t buf = 0;
	
	while (count--) {
	  // wait for start bit
	  do {
	    UARTWaitForFallingClk;
	  } while (bit_test1(UART_DATA_PORT));
	  
	  for (bitcount=8; bitcount > 0; bitcount--) {
	    // read next bit after falling clock edge
	    buf >>= 1;
	    UARTWaitForFallingClk;
	    if (bit_test1(UART_DATA_PORT)) buf |= 0x80;
	  } // for i
	  
	  // wait for stop bit
	  UARTWaitForFallingClk;
	  if (bit_test1(UART_DATA_PORT)) {
	    // invalid stop bit (should be 0)
	    return readbytes;
	  } // if

	  // store value
	  *val = buf;
	  val++;
	  readbytes++;
	} // while
	
	return readbytes;
}
